"""
OAuth2-based client for more advanced applications.

Single-user CLI
---------------

This method is suitable for testing or where indefinite authentication is
required, for example a long-running service.
In general, regular single-user applications should instead take the user's
personal API key and use the simple client.

.. code:: python

    # Initialise client with basic defaults
    from srf_client import *
    srf = SRFDataOAuth(client_id='...', client_secret='...')

    # Start a browser login session
    import webbrowser
    webbrowser.open_new_tab(srf.start_login())

    # Receive auth code
    code = input('Enter authorization code: ')
    srf.login_with_code(code)

    # Test connection
    srf.ping()

Multi-user webapp
-----------------

Each user must use their own separate authenticated session.
This can be achieved with multiple client instances, or by switching the token
of a single instance before each request.

This is a basic flask example. A robust app will need more state and error
handling.

.. code:: python

    from flask import Flask, jsonify, redirect, request, session
    from srf_client.oauth import *

    app = Flask(__name__)
    srf_config = {'client_id': '...',
                  'client_secret': '...',
                  'redirect_uri': 'https://example.com/oauth_callback'}

    def save_oauth_token(token):
        session['oauth_token'] = token

    def get_client():
        srf_client = SRFDataOAuth(save_token=save_oauth_token, **srf_config)
        if 'oauth_token' in session:
            srf_client.login_with_token(session['oauth_token'])
        return srf_client

    @app.route('/login')
    def login():
        return redirect(get_client().start_login())

    @app.route('/oauth_callback')
    def callback():
        get_client().login_with_callback(request.url)
        return redirect('/')

    @app.route('/logout')
    def logout():
        get_client().logout()
        session.pop('oauth_token')
        return redirect('/')

    @app.route('/')
    def index():
        return jsonify(get_client().ping())

Client-only access
------------------
.. versionadded:: 1.1

Some APIs provide access to shared data and do not require user authorization.
If your app only uses these then configure a single scope of :code:`CLIENT`.
Saving the token is not necessary, as a new one will be fetched whenever
needed.

.. code:: python

    from srf_client import *

    srf = SRFDataOAuth(client_id='...', client_secret='...', scope=('CLIENT',))
    srf.login_with_client()
    srf.ping()

"""

from typing import Callable, Mapping, Sequence
from oauthlib.oauth2 import BackendApplicationClient, OAuth2Error, \
    WebApplicationClient
from requests import RequestException
from requests_oauthlib import OAuth2Session

from .client import SRFData

__all__ = ['SRFDataOAuth']


class SRFDataOAuth(SRFData):
    """SRF Data client. Provides OAuth2-based access to the REST API."""

    def __init__(self, client_id: str, client_secret: str,
                 scope: Sequence[str] = ('USER',),
                 redirect_uri='https://data.csrf.ac.uk/oauth/callback',
                 save_token: Callable[[Mapping], None] = None,
                 auth_endpoint='https://data.csrf.ac.uk/oauth/authorize',
                 token_endpoint='https://data.csrf.ac.uk/oauth/token',
                 revoke_endpoint='https://data.csrf.ac.uk/oauth/revoke',
                 **kwargs):
        """
        Initialise a new instance.

        :param client_id: Registered client ID
        :param client_secret: Registered client secret
        :param scope: (optional) Request additional authorisation
        :param redirect_uri: (optional) Callback URL to receive auth result
        :param save_token: (optional) Callback function to save token state
            e.g. in a web session
        :param root: (optional) API root URL
        :param auth_endpoint: (optional) Override OAuth URL
        :param token_endpoint: (optional) Override OAuth URL
        :param revoke_endpoint: (optional) Override OAuth URL
        """
        super().__init__(client_id=client_id,
                         scope=scope,
                         redirect_uri=redirect_uri,
                         token_endpoint=token_endpoint,
                         **kwargs)
        self._client_id = client_id
        self._client_secret = client_secret
        self._scope = tuple(scope)
        self._redirect_uri = redirect_uri
        self._auth_endpoint = auth_endpoint
        self._token_endpoint = token_endpoint
        self._revoke_endpoint = revoke_endpoint
        self._save_token = save_token

    def _make_session(self, **kwargs):
        # the client class determines the grant_type
        if tuple(kwargs['scope']) == ('CLIENT',):
            # client_credentials
            client = BackendApplicationClient(kwargs['client_id'],
                                              scope=kwargs['scope'])
        else:
            # authorization_code
            client = WebApplicationClient(kwargs['client_id'],
                                          scope=kwargs['scope'],
                                          redirect_url=kwargs['redirect_uri'])

        session = OAuth2Session(
            client=client,
            scope=kwargs['scope'],
            redirect_uri=kwargs['redirect_uri'],
            auto_refresh_url=kwargs['token_endpoint'],
            token_updater=self._update_token
        )
        session.request = self._wrap_request(session.request)
        session.refresh_token = self._wrap_refresh(session.refresh_token)
        self._session = session

    def _update_token(self, token):
        self._session.token = token
        if self._save_token is not None:
            self._save_token(token)

    def _wrap_request(self, func):
        def wrapper(*args, **kwargs):
            try:
                response = func(*args, **kwargs)
                if response.status_code == 401:
                    self._clear_token()
                return response
            except OAuth2Error as exc:
                self._clear_token()
                raise RequestException(exc.error) from exc
        return wrapper

    def _wrap_refresh(self, func):
        # https://github.com/requests/requests-oauthlib/issues/264
        def wrapper(*args, **kwargs):
            if self._scope == ('CLIENT',):
                self.login_with_client()
                return self._session.token
            else:
                kwargs['auth'] = (self._client_id, self._client_secret)
                kwargs.pop('allow_redirects', None)
                return func(*args, **kwargs)
        return wrapper

    def _clear_token(self):
        self._session.token = {}
        self._session.access_token = None

    @property
    def logged_in(self) -> bool:
        """
        Whether the client is currently authenticated.

        Does not make a remote call to check for e.g. expiration.
        Use ``ping()`` for an active check.
        """
        return self._session.authorized

    def start_login(self) -> str:
        """
        Begin an interactive login session.

        Complete with either ``login_with_callback`` or ``login_with_code``.

        :return: The URL that the user should visit.
        """
        return self._session.authorization_url(self._auth_endpoint)[0]

    def login_with_callback(self, response_callback: str) -> None:
        """
        Complete a login session using a received callback URL.

        :param response_callback: Full callback URL
        """
        self._session.fetch_token(
            token_url=self._token_endpoint,
            client_secret=self._client_secret,
            authorization_response=response_callback
        )
        if self._save_token is not None:
            self._save_token(self._session.token)

    def login_with_code(self, authorization_code: str) -> None:
        """
        Complete a login session using a hand-entered code.

        :param authorization_code: User's auth code
        """
        self._session.fetch_token(
            token_url=self._token_endpoint,
            client_secret=self._client_secret,
            code=authorization_code
        )
        if self._save_token is not None:
            self._save_token(self._session.token)

    def login_with_client(self) -> None:
        """
        Login as a non-user client.

        .. versionadded:: 1.1

        """
        self._session.fetch_token(
            token_url=self._token_endpoint,
            client_secret=self._client_secret
        )
        if self._save_token is not None:
            self._save_token(self._session.token)

    def login_with_token(self, token: Mapping) -> None:
        """
        Restore auth state from a saved token.

        :param token: object previously received via ``save_token``
        """
        self._session.token = token

    def logout(self) -> None:
        """
        Invalidate the current auth state.

        Saved tokens will not be reusable.
        """
        token = self._session.token
        if token:
            oauth_client = self._session._client  # noqa
            url, headers, body = oauth_client.prepare_token_revocation_request(
                self._revoke_endpoint,
                token.get('refresh_token', token.get('access_token')),
                token_type_hint=None)
            self._session.post(url, headers=headers, data=body,
                               auth=(self._client_id, self._client_secret))
        self._clear_token()



from datetime import datetime
from typing import Optional

import iso8601
from geopy import Point
from geopy.distance import Distance, geodesic


def parse_point(data) -> Optional[Point]:
    """Parse returned data into a ``Point`` object."""
    if data is None:
        return None
    else:
        return Point(data.get("lat"), data.get("lon"), data.get("alt"))


def parse_distance(value, unit='kilometers') -> Optional[Distance]:
    """Parse returned data into a ``Distance`` object."""
    if value is None:
        return None
    else:
        return geodesic(**{unit: value})


def parse_datetime(value) -> Optional[datetime]:
    """Parse returned data into a ``datetime`` object."""
    if value is None:
        return None
    else:
        return iso8601.parse_date(value)



from .model import Trailer, Trial, Trip, Vehicle
from .paging import Page, PagingClient
from .util import parse_datetime, parse_distance


class TripsClient(PagingClient[Trip]):
    """
    Access to trips API.

    Acquire an instance via ``SRFData.trips``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'trips')

    def _parse_obj(self, data, uri) -> Trip:
        return Trip(
            client=self._client,
            uri=uri,
            source=data['source'],
            types=frozenset(data['types']),
            start_time=parse_datetime(data.get('startTime')),
            end_time=parse_datetime(data.get('endTime')),
            consumption=data.get('consumption'),
            distance=parse_distance(data.get('distance')),
            trailer_uri=data['trailer']['_location'],
            trial_uri=data['trial']['_location'],
            vehicle_uri=data['vehicle']['_location']
        )

    def find_all(self, lazy=False, **kwargs) -> Page[Trip]:
        """
        Find all recorded trips.

        :param lazy: Defer fetching of the Trip objects
        :param kwargs: Additional field filters
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/trips', params)

    def find_by_vehicle(self, vehicle: Vehicle, lazy=False, **kwargs) \
            -> Page[Trip]:
        """
        Find all trips for the given vehicle.

        :param vehicle: Parent vehicle object
        :param lazy: Defer fetching of the Trip objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page(vehicle.uri + '/trips', params)

    def find_by_trial(self, trial: Trial, lazy=False, **kwargs) -> Page[Trip]:
        """
        Find all trips for the given trial.

        :param trial: Parent trial object
        :param lazy: Defer fetching of the Trip objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page(trial.uri + '/trips', params)

    def find_by_trailer(self, trailer: Trailer, lazy=False, **kwargs) \
            -> Page[Trip]:
        """
        Find all trips for the given trailer.

        :param trailer: Parent trailer object
        :param lazy: Defer fetching of the Trip objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page(trailer.uri + '/trips', params)



"""
Use with listing methods to filter the results.

    .. code:: python

        srf.trips.find_all(distance=between(0, 50))

Related entities can be traversed, which required using dict syntax:

    .. code:: python

        srf.trips.find_all(**{vehicle.make=eq('Scania')})

.. warning::
    Using unsupported operators for a field type will give unexpected
    results.
"""

from abc import ABC
from datetime import datetime
from math import isinf

__all__ = [
    'Equals', 'Prefix', 'Between', 'Contains', 'AnyOf', 'Near',
    'eq', 'prefix', 'between', 'contains', 'any_of', 'near'
]

from geopy import Point
from geopy.distance import Distance


class Operator(ABC):
    def __init__(self, *values):
        self.values = [str(v) for v in values]

    def __iter__(self):
        return iter(self.values)


class Equals(Operator):
    """Field equals value."""

    def __init__(self, value):
        super().__init__(value)


class Prefix(Operator):
    """String field starts with value."""

    def __init__(self, value):
        super().__init__(str(value) + '*')


class Between(Operator):
    """Field is between two values (inclusive)."""

    def __init__(self, v1, v2):
        super().__init__(self._convert(v1), self._convert(v2))

    @staticmethod
    def _convert(value):
        if isinstance(value, datetime):
            return value.isoformat(timespec='milliseconds')
        elif isinf(value):
            return 'Infinity' if value > 0 else '-Infinity'
        else:
            return value


class Contains(Operator):
    """Array field contains value(s)."""


class AnyOf(Operator):
    """String field matches any of these operators."""

    def __init__(self, *args: Operator):
        super().__init__(*(v for op in args for v in op.values))


class Near(Operator):
    """Point is within region centred on point."""

    def __init__(self, point: Point, distance: Distance):
        super().__init__(f'({point.latitude},{point.longitude}),{distance.km}')


eq = Equals
prefix = Prefix
between = Between
contains = Contains
any_of = AnyOf
near = Near



from .model import Trailer
from .paging import Page, PagingClient


class TrailersClient(PagingClient[Trailer]):
    """
    Access to trailers API.

    Acquire an instance via ``SRFData.trailers``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'trailers')

    def _parse_obj(self, data, uri) -> Trailer:
        return Trailer(
            client=self._client,
            uri=uri,
            serial=data['serial'],
            length=data.get('length'),
            make=data.get('make'),
            model=data.get('model'),
            type=data.get('type'),
            axles=data.get('axles'),
            tires_per_axle=data.get('tiresPerAxle'),
            description=data.get('description'),
            org_uri=data['organisation']['_location']
        )

    def find_all(self, lazy=False, **kwargs) -> Page[Trailer]:
        """
        Find all known trailers.

        :param lazy: Defer fetching of the Vehicle objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/trailers', params)



"""
Can be passed with field filters to control the sort order.

    .. code:: python

        srf.organisations.find_all(sort=asc('name'))

"""

from abc import ABC

__all__ = [
    'Ascending', 'Descending',
    'asc', 'desc'
]


class Sort(ABC):
    def __init__(self, field: str, direction: str):
        self.field = field
        self.direction = direction

    def __str__(self):
        return f'{self.field},{self.direction}'


class Ascending(Sort):
    """
    Ascending sort order.

    :param field: Field to sort by
    """

    def __init__(self, field: str):
        super().__init__(field, 'asc')


class Descending(Sort):
    """
    Descending sort order.

    :param field: Field to sort by
    """

    def __init__(self, field: str):
        super().__init__(field, 'desc')


asc = Ascending
desc = Descending



from .model import Location, Route, VehicleClass
from .paging import Page, PagingClient
from .util import parse_datetime, parse_distance


class RoutesClient(PagingClient[Route]):
    """
    Access to Routes API.

    Acquire an instance via ``SRFData.routes``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'routes')

    def _parse_obj(self, data, uri) -> Route:
        return Route(
            client=self._client,
            uri=uri,
            org_uri=data['organisation']['_location'],
            start_location_uri=data['startLocation']['_location'],
            end_location_uri=data['endLocation']['_location'],
            depot_uri=data['depot']['_location'],
            vehicle_class_uri=data['vehicleClass']['_location'],
            plan=data['plan'],
            start_time=parse_datetime(data['startTime']),
            end_time=parse_datetime(data['endTime']),
            orders=data['orders'],
            shift_start=parse_datetime(data.get('shiftStart')),
            shift_end=parse_datetime(data.get('shiftEnd')),
            distance=parse_distance(data.get('distance')),
            registration=data.get('registration'),
            total_units=data.get('totalUnits'),
            total_weight=data.get('totalWeight')
        )

    def find_all(self, lazy=False, **kwargs) -> Page[Route]:
        """
        Find all known routes.

        :param lazy: Defer fetching of the Route objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/routes', params)

    def find_by_location(self, location: Location, lazy=False, **kwargs) \
            -> Page[Route]:
        """
        Find all routes involving the given location.

        :param location: Parent location object
        :param lazy: Defer fetching of the Route objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page(location.uri + '/routes', params)

    def find_by_vehicle_class(self, vehicle_class: VehicleClass, lazy=False,
                              **kwargs) -> Page[Route]:
        """
        Find all routes using the given vehicle class.

        :param vehicle_class: Parent vehicle class object
        :param lazy: Defer fetching of the Route objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page(vehicle_class.uri + '/routes', params)



from .model import Organisation
from .paging import Page, PagingClient


class OrganisationsClient(PagingClient[Organisation]):
    """
    Access to Organisations API.

    Acquire an instance via ``SRFData.organisations``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'organisations')

    def _parse_obj(self, data, uri) -> Organisation:
        return Organisation(
            client=self._client,
            uri=uri,
            name=data['name']
        )

    def find_all(self, lazy=False, **kwargs) -> Page[Organisation]:
        """
        Find all known organisations.

        :param lazy: Defer fetching of the Organisation objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/organisations', params)



"""Data model for the API."""
from collections import namedtuple

import attr  # may replace with @dataclass in py37
from datetime import datetime
from geopy import Point
from geopy.distance import Distance
from immutabledict import immutabledict
from itertools import chain
from typing import Iterable, Mapping, Optional, Sequence

from srf_client.paging import paged_items
from . import client, filter as f, paging

__all__ = [
    'Calibration', 'DataType', 'ExternalVehicleClass', 'FieldDef', 'Fleet',
    'FleetEntry', 'Leg', 'Location', 'Measurement', 'Order', 'Organisation',
    'Route', 'Trailer', 'Trial', 'Trip', 'Vehicle', 'VehicleClass'
]


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Organisation:
    """
    An affiliated or former organisation.

    :param uri: URI for this organisation
    :param name: Name of the organisation
    """

    _client: 'client.SRFData'
    uri: str
    name: str

    def get_fleets(self, lazy=False) -> 'paging.Page[Fleet]':
        """
        Fetch fleets for organisation.

        :param lazy: Defer fetching of the Fleet objects
        """
        return self._client.fleets.find_all(lazy=lazy, **{
            'organisation.name': f.eq(self.name)
        })

    def get_routes(self, lazy=False) -> 'paging.Page[Route]':
        """
        Fetch routes for organisation.

        :param lazy: Defer fetching of the Route objects
        """
        return self._client.routes.find_all(lazy=lazy, **{
            'organisation.name': f.eq(self.name)
        })

    def get_trailers(self, lazy=False) -> 'paging.Page[Trailer]':
        """
        Fetch trailers for organisation.

        :param lazy: Defer fetching of the Trailer objects
        """
        return self._client.trailers.find_all(lazy=lazy, **{
            'organisation.name': f.eq(self.name)
        })

    def get_trials(self, lazy=False) -> 'paging.Page[Trial]':
        """
        Fetch trials for organisation.

        :param lazy: Defer fetching of the Trial objects
        """
        return self._client.trials.find_all(lazy=lazy, **{
            'organisation.name': f.eq(self.name)
        })

    def get_vehicle_classes(self, lazy=False) -> 'paging.Page[VehicleClass]':
        """
        Fetch vehicle classes for organisation.

        :param lazy: Defer fetching of the VehicleClass objects
        """
        return self._client.vehicle_classes.find_all(lazy=lazy, **{
            'organisation.name': f.eq(self.name)
        })

    def get_vehicles(self, lazy=False) -> 'paging.Page[Vehicle]':
        """
        Fetch vehicles for organisation.

        :param lazy: Defer fetching of the Vehicle objects
        """
        return self._client.vehicles.find_all(lazy=lazy, **{
            'organisation.name': f.eq(self.name)
        })


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Trial:
    """
    A specific project investigation.

    :param uri: URI for this trial
    :param description: Description of the trial
    :param active: Whether data collection is ongoing
    """

    _client: 'client.SRFData'
    _org_uri: str
    uri: str
    description: str
    active: bool

    @property
    def organisation(self) -> 'Organisation':  # noqa: D102
        return self._client.organisations.get(uri=self._org_uri)

    def get_trips(self, lazy=False, **kwargs) -> 'paging.Page[Trip]':
        """
        Fetch trips for trial.

        :param lazy: Defer fetching of the Trip objects
        :param kwargs: Additional field filters
        """
        return self._client.trips.find_by_trial(self, lazy=lazy, **kwargs)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Vehicle:
    """
    A motor vehicle.

    :param uri: URI for this vehicle
    :param registration: Vehicle registration (number plate)
    :param vin: Vehicle identification number (ISO 3779)
    :param calibrated: Whether the app was calibrated on installation.
        If missing then data types 11 and 12 are in an arbitrary basis.
    :param make: Vehicle manufacturer
    :param model: Manufacturer’s model designation
    :param type: Vehicle classification (e.g. ``RIGID``, ``ARTIC``, ``VAN``,
        ``BUS``, ``CAR``)
    :param fuel: Fuel type (e.g. ``DIESEL``, ``PETROL``, ``CNG``, ``LNG``,
        ``H2``, ``ELECTRIC``)
    :param weight_class: Max weight class (tonnes)
    :param euro_standard: European emission standard
    :param description: Additional description/notes
    """

    _client: 'client.SRFData'
    _org_uri: str
    uri: str
    registration: Optional[str] = None
    vin: Optional[str] = None
    calibrated: bool = False
    make: Optional[str] = None
    model: Optional[str] = None
    type: Optional[str] = None
    fuel: Optional[str] = None
    weight_class: Optional[float] = None
    euro_standard: Optional[int] = None
    description: Optional[str] = None

    @property
    def organisation(self) -> 'Organisation':  # noqa: D102
        return self._client.organisations.get(uri=self._org_uri)

    def get_trips(self, lazy=False, **kwargs) -> 'paging.Page[Trip]':
        """
        Fetch trips for vehicle.

        :param lazy: Defer fetching of the Trip objects
        :param kwargs: Additional field filters
        """
        return self._client.trips.find_by_vehicle(self, lazy=lazy, **kwargs)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Trailer:
    """
    A towed vehicle or container.

    :param uri: URI for this trailer
    :param serial: Trailer serial number
    :param length: Length of trailer chassis (m)
    :param make: Trailer manufacturer
    :param model: Manufacturer’s model designation
    :param type: Trailer classification (e.g. ``SINGLE``, ``HIGH``, ``DOUBLE``,
        ``TEARDROP``, ``LONG``, ``SLOPED``, ``TANK``)
    :param axles: Number of axles
    :param tires_per_axle: Number of tires per axle
    :param description: Additional description/notes
    """

    _client: 'client.SRFData'
    _org_uri: str
    uri: str
    serial: str
    length: Optional[float] = None
    make: Optional[str] = None
    model: Optional[str] = None
    type: Optional[str] = None
    axles: Optional[int] = None
    tires_per_axle: Optional[int] = None
    description: Optional[str] = None

    @property
    def organisation(self) -> 'Organisation':  # noqa: D102
        return self._client.organisations.get(uri=self._org_uri)

    def get_trips(self, lazy=False, **kwargs) -> 'paging.Page[Trip]':
        """
        Fetch trips for trailer.

        :param lazy: Defer fetching of the Trip objects
        :param kwargs: Additional field filters
        """
        return self._client.trips.find_by_trailer(self, lazy=lazy, **kwargs)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Trip:
    """
    A single journey.

    Can be defined manually, or as all the legs that begin on a particular day.

    :param uri: URI for this trip
    :param types: Set of available data types
    :param start_time: First recorded timestamp
    :param end_time: Last recorded timestamp
    :param consumption: Fuel usage over the trip (litres)
    :param distance: Distanced travelled over the trip
    """

    _client: 'client.SRFData'
    _trailer_uri: str
    _trial_uri: str
    _vehicle_uri: str
    uri: str
    source: str
    types: frozenset = frozenset()
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None
    consumption: Optional[float] = None
    distance: Optional[Distance] = None

    @property
    def trailer(self) -> 'Trailer':  # noqa: D102
        return self._client.trailers.get(uri=self._trailer_uri)

    @property
    def trial(self) -> 'Trial':  # noqa: D102
        return self._client.trials.get(uri=self._trial_uri)

    @property
    def vehicle(self) -> 'Vehicle':  # noqa: D102
        return self._client.vehicles.get(uri=self._vehicle_uri)

    def get_legs(self, lazy=False, **kwargs) -> 'paging.Page[Leg]':
        """
        Fetch legs for trip.

        :param lazy: Defer fetching of the Leg objects
        :param kwargs: Additional field filters
        """
        return self._client.legs.find_by_trip(self, lazy=lazy, **kwargs)

    def get_data(self, include=None) -> Iterable['Measurement']:
        """
        Return available measurements.

        :param include: Data types to include. By default all are included.
        """
        return chain.from_iterable(leg.get_data(include=include)
                                   for leg in paged_items(self.get_legs()))


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Leg:
    """
    A largely arbitrary section of a vehicle’s journey.

    Remaining stopped for an amount of time triggers the end of a leg.

    :param uri: URI for this leg
    :param trip: Parent trip
    :param start_time: First recorded timestamp
    :param end_time: Last recorded timestamp
    :param types: Set of available data types
    :param start_location: Location at start of leg
    :param end_location: Location at end of leg
    :param start_distance: Odometer reading at start of leg
    :param end_distance: Odometer reading at end of leg
    :param consumption: Fuel usage over the leg (litres)
    :param fuel_level: Fuel level at end of leg (percent)
    :param calibration: Calibration data
    :param weight: Gross vehicle weight during the leg (kilograms)
    :param weight_source: Source of the weight value
    """

    _client: 'client.SRFData'
    _trip_uri: str
    uri: str
    start_time: datetime
    end_time: datetime
    types: frozenset = frozenset()
    start_location: Optional[Point] = None
    end_location: Optional[Point] = None
    start_distance: Optional[Distance] = None
    end_distance: Optional[Distance] = None
    consumption: Optional[float] = None
    fuel_level: Optional[float] = None
    weight: Optional[int] = None
    weight_source: Optional[str] = None

    @property
    def trip(self) -> 'Trip':  # noqa: D102
        return self._client.trips.get(uri=self._trip_uri)

    @property
    def calibration(self) -> 'Calibration':  # noqa: D102
        return self._client.legs.get_calibration(self)

    def get_data(self, include=None) -> Iterable['Measurement']:
        """
        Return available measurements.

        :param include: Data types to include. By default all are included.
        """
        return self._client.legs.get_data(self, include=include)

    def get_raw_data(self, include=None) -> Iterable['Measurement']:
        """
        Return the raw measurements.

        :param include: Data types to include. By default all are included.
        """
        return self._client.legs.get_raw_data(self, include=include)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Calibration:
    """
    Leg calibration data for vehicle dynamics.

    :param forward: Unit vector pointing in the direction of forward travel
    :param left: Lateral unit vector (forward × up)
    :param up: Unit vector pointing against the action of gravity
    :param inv_gravity: Inverse gravitational acceleration in device's basis
    """

    forward: tuple
    left: tuple
    up: tuple
    inv_gravity: tuple


Measurement = namedtuple('Measurement', ['timestamp', 'type', 'data'])
DataType = namedtuple('DataType', ['type', 'description', 'fields'])
FieldDef = namedtuple('FieldDef', ['name', 'unit'])


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Fleet:
    """
    Set of vehicles available for operations.

    :param uri: URI for this fleet
    :param organisation: Owning organisation
    :param name: Name of this fleet
    :param entries: Fleet composition
    """

    _client: 'client.SRFData'
    _org_uri: str
    uri: str
    name: str
    entries: Sequence['FleetEntry']

    @property
    def organisation(self) -> Organisation:  # noqa: D102
        return self._client.organisations.get(uri=self._org_uri)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class FleetEntry:
    """
    Component of a fleet.

    :param location: Vehicle home location
    :param vehicle_class: Vehicle description
    :param count: Number of vehicles of this type at this location
    """

    _client: 'client.SRFData'
    _location_uri: str
    _vehicle_class_uri: str
    count: int

    @property
    def location(self) -> 'Location':  # noqa: D102
        return self._client.locations.get(uri=self._location_uri)

    @property
    def vehicle_class(self) -> 'VehicleClass':  # noqa: D102
        return self._client.vehicle_classes.get(uri=self._vehicle_class_uri)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Location:
    """
    Significant location.

    :param uri: URI for this location
    :param organisation: Owning organisation
    :param name: Name for this location
    :param post_code: Location's postal code
    :param point: Coordinates of the location
    """

    _client: 'client.SRFData'
    _org_uri: str
    uri: str
    name: Optional[str]
    post_code: Optional[str]
    point: Optional[Point]

    @property
    def organisation(self) -> 'Organisation':  # noqa: D102
        return self._client.organisations.get(uri=self._org_uri)

    def get_routes(self, lazy=False, **kwargs) -> 'paging.Page[Route]':
        """
        Fetch routes for location.

        :param lazy: Defer fetching of the Leg objects
        :param kwargs: Additional field filters
        """
        return self._client.routes.find_by_location(self, lazy, **kwargs)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Route:
    """
    A logistics operation route, planned or completed.

    :param uri: URI for this route
    :param organisation: Owning organisation
    :param start_location: Start of the route
    :param end_location: End of the route
    :param depot: Operational base location
    :param vehicle_class: Type of vehicle used
    :param plan: Whether this route describes a planned or an actual route
    :param start_time: Start time
    :param end_time: End time
    :param distance: Total distance travelled
    :param shift_start: Start of driver's shift
    :param shift_end: End of driver's shift
    :param registration: Specific vehicle that was used
    :param orders: Total number of orders involved
    :param total_units: Total number of units moved
    :param total_weight: Total weight of goods moved (kilograms)
    """

    _client: 'client.SRFData'
    _org_uri: str
    _start_location_uri: str
    _end_location_uri: str
    _depot_uri: str
    _vehicle_class_uri: str
    uri: str
    plan: bool
    start_time: datetime
    end_time: datetime
    distance: Optional[Distance] = None
    shift_start: Optional[datetime] = None
    shift_end: Optional[datetime] = None
    registration: Optional[str] = None
    orders: int = 0
    total_units: Optional[float] = None
    total_weight: Optional[float] = None

    @property
    def organisation(self) -> 'Organisation':  # noqa: D102
        return self._client.organisations.get(uri=self._org_uri)

    @property
    def start_location(self) -> 'Location':  # noqa: D102
        return self._client.locations.get(uri=self._start_location_uri)

    @property
    def end_location(self) -> 'Location':  # noqa: D102
        return self._client.locations.get(uri=self._end_location_uri)

    @property
    def depot(self) -> 'Location':  # noqa: D102
        return self._client.locations.get(uri=self._depot_uri)

    @property
    def vehicle_class(self) -> 'VehicleClass':  # noqa: D102
        return self._client.vehicle_classes.get(uri=self._vehicle_class_uri)

    def get_orders(self, lazy=False, **kwargs) -> 'paging.Page[Order]':
        """
        Fetch orders for route.

        :param lazy: Defer fetching of the Leg objects
        :param kwargs: Additional field filters
        """
        return self._client.orders.find_by_route(self, lazy=lazy, **kwargs)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class Order:
    """
    Transported cargo.

    :param uri: URI for this order
    :param route: Route transporting this cargo
    :param pickup: Postal code of pickup location
    :param destination: Postal code of destination location
    :param time_window_start: Start of the delivery window
    :param time_window_end: End of the delivery window
    :param pickup_arrive: Arrival at pickup location
    :param pickup_depart: Departure from pickup location
    :param destination_arrive: Arrival at destination
    :param destination_depart: Departure from destination
    :param distance: Distance travelled for delivery
    :param ambient: Number of ambient units
    :param chilled: Number of chilled units
    :param frozen: Number of frozen units
    :param weight: Weight of order (kilograms)
    """

    _client: 'client.SRFData'
    _route_uri: str
    uri: str
    pickup: str
    destination: str
    time_window_start: Optional[datetime] = None
    time_window_end: Optional[datetime] = None
    pickup_arrive: Optional[datetime] = None
    pickup_depart: Optional[datetime] = None
    destination_arrive: Optional[datetime] = None
    destination_depart: Optional[datetime] = None
    distance: Optional[Distance] = None
    ambient: Optional[float] = None
    chilled: Optional[float] = None
    frozen: Optional[float] = None
    weight: Optional[float] = None

    @property
    def route(self) -> 'Route':  # noqa: D102
        return self._client.routes.get(uri=self._route_uri)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class VehicleClass:
    """
    A specific make/model/configuration of road vehicle.

    :param uri: URI for this vehicle class
    :param organisation: Owning organisation
    :param description: Additional description/notes
    :param make: Vehicle manufacturer
    :param model: Manufacturer’s model designation
    :param type: Vehicle classification (e.g. ``RIGID``, ``ARTIC``, ``VAN``,
        ``BUS``, ``CAR``)
    :param fuel: Fuel type (e.g. ``DIESEL``, ``PETROL``, ``CNG``, ``LNG``,
        ``H2``, ``ELECTRIC``)
    :param fuel_capacity: Full fuel capacity (L or kg or kWh)
    :param fuel_efficiency: Fuel usage over distance (L/km or kg/km or kWh/km)
    :param weight_class: Max weight class (tonnes)
    :param euro_standard: European emission standard
    :param drive_power: Power output of the drive system (kilowatts)
    :param fridge_power: Power output of the cargo cooling system (kilowatts)
    :param limits: Various numeric limits for constraint problems
    """

    _client: 'client.SRFData'
    _org_uri: str
    uri: str
    description: Optional[str] = None
    make: Optional[str] = None
    model: Optional[str] = None
    type: Optional[str] = None
    fuel: Optional[str] = None
    fuel_capacity: Optional[float] = None
    fuel_efficiency: Optional[float] = None
    weight_class: Optional[float] = None
    euro_standard: Optional[int] = None
    drive_power: Optional[float] = None
    fridge_power: Optional[float] = None
    limits: Mapping[str, float] = immutabledict()

    @property
    def organisation(self) -> 'Organisation':  # noqa: D102
        return self._client.organisations.get(uri=self._org_uri)

    def get_routes(self, lazy=False, **kwargs) -> 'paging.Page[Route]':
        """
        Fetch routes for vehicle class.

        :param lazy: Defer fetching of the Leg objects
        :param kwargs: Additional field filters
        """
        return self._client.routes.find_by_vehicle_class(self, lazy, **kwargs)


@attr.s(frozen=True, slots=True, auto_attribs=True, kw_only=True)
class ExternalVehicleClass:
    """
    A specific make/model/configuration of vehicle from an external system.

    See :class:`~.VehicleClass`.
    """

    description: Optional[str] = None
    make: Optional[str] = None
    model: Optional[str] = None
    type: Optional[str] = None
    fuel: Optional[str] = None
    fuel_capacity: Optional[float] = None
    fuel_efficiency: Optional[float] = None
    weight_class: Optional[float] = None
    euro_standard: Optional[int] = None
    drive_power: Optional[float] = None
    fridge_power: Optional[float] = None



import pandas as pd
from datetime import timedelta
from functools import partial, reduce
from typing import Any, Callable, Collection, List, Optional, Union

from .model import Leg, Trip

to_numeric = partial(pd.to_numeric, errors='coerce')


def _identity(x):
    return x


def get_data_frame(self: Union[Leg, Trip],
                   data_types: Union[str, Collection[str]],
                   resolution: Union[str, timedelta, None] = None,
                   conversion: Optional[Callable[[str], Any]] = to_numeric
                   ) -> pd.DataFrame:
    """
    Return available measurements as a time-series :code:`DataTable`.

    :code:`Series` will be named :code:`{TYPE}_{FIELD}` in the order given
    by :code:`data_types`.

    Data will be interpolated to the specified resolution.

    :param data_types: Data types to fetch.
    :param resolution: Target resolution. If :code:`None` (the default) then
        it will use the approximate resolution of the first requested type.
    :param conversion: Conversion to apply to all values. If :code:`None`
        then all values will be strings.
    """
    if isinstance(data_types, str):
        data_types = [data_types]

    if conversion is None:
        conversion = _identity

    collected: List[Any] = [([], []) for _ in data_types]
    data_type_idx = {data_type: idx
                     for idx, data_type in enumerate(data_types)}

    for m in self.get_data(include=data_types):
        idx = data_type_idx[m.type]
        collected[idx][0].append(pd.to_datetime(m.timestamp, unit='ms'))
        collected[idx][1].append(tuple(conversion(v)
                                       for v in m.data.split(',')))

    if resolution is None and len(data_types) > 1:
        try:
            resolution = collected[0][0][1] - collected[0][0][0]
        except IndexError:
            raise ValueError('No data for first type and no resolution given')

    new_index = None
    if resolution is not None:
        start = pd.to_datetime(self.start_time).ceil(resolution)
        end = pd.to_datetime(self.end_time).ceil(resolution)
        new_index = pd.date_range(start=start, end=end, freq=resolution)
        # np.datetime64 cannot hold timezone
        new_index = new_index.tz_localize(None)

    type_defs = self._client.get_types()

    for idx, data_type in enumerate(data_types):
        if len(collected[idx][0]) == 0:
            collected[idx] = pd.DataFrame()
            continue

        index = pd.DatetimeIndex(collected[idx][0])
        data = collected[idx][1]
        columns = ['{} {}'.format(data_type, field.name)
                   for field in type_defs[data_type].fields]
        collected[idx] = pd.DataFrame.from_records(data=data,
                                                   index=index,
                                                   columns=columns)
        if new_index is not None:
            if not index.is_unique:
                # https://csrf.atlassian.net/browse/PLAT-222
                collected[idx] = collected[idx][~index.duplicated()]
                index = collected[idx].index

            collected[idx].flags.allows_duplicate_labels = False
            collected[idx] = collected[idx] \
                .reindex(index.union(new_index)) \
                .interpolate(method='index') \
                .reindex(new_index)

    collected = [df for df in collected if not df.empty]
    if not collected:
        return pd.DataFrame()
    else:
        return reduce(pd.DataFrame.join, collected)


Leg.get_data_frame = get_data_frame
Trip.get_data_frame = get_data_frame



from typing import Collection, Iterable, Union

from .model import Calibration, Leg, Measurement, Trip
from .paging import Page, PagingClient
from .util import parse_datetime, parse_distance, parse_point


class LegsClient(PagingClient[Leg]):
    """
    Access to legs API.

    Acquire an instance via ``SRFData.trips``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'legs')

    def _parse_obj(self, data, uri) -> Leg:
        return Leg(
            client=self._client,
            uri=uri,
            trip_uri=data['trip']['_location'],
            start_time=parse_datetime(data['startTime']),
            end_time=parse_datetime(data['endTime']),
            types=frozenset(data['types']),
            start_location=parse_point(data.get('startLocation')),
            end_location=parse_point(data.get('endLocation')),
            start_distance=parse_distance(data.get('startDistance')),
            end_distance=parse_distance(data.get('endDistance')),
            consumption=data.get('consumption'),
            fuel_level=data.get('fuelLevel'),
            weight=data.get('weight'),
            weight_source=data.get('weightSource')
        )

    def find_all(self, lazy=False, **kwargs) -> Page[Leg]:
        """
        Find all recorded legs.

        :param lazy: Defer fetching of the Leg objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/legs', params)

    def find_by_trip(self, trip: Trip, lazy=False, **kwargs) -> Page[Leg]:
        """
        Find all legs for the given trip.

        :param trip: Parent trip object
        :param lazy: Defer fetching of the Leg objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page(trip.uri + '/legs', params)

    def get_calibration(self, leg: Leg) -> Calibration:
        """Return the raw calibration data for a given leg."""
        response = self._client.get(leg.uri + '/data', stream=True)
        response.raise_for_status()
        header = next(response.iter_lines(decode_unicode=True))
        parts = header.split(',')
        return Calibration(
            forward=tuple(float(n) for n in parts[5].split(':')),
            left=tuple(float(n) for n in parts[6].split(':')),
            up=tuple(float(n) for n in parts[7].split(':')),
            inv_gravity=tuple(float(n) for n in parts[8].split(':'))
        )

    def get_data(
        self,
        leg: Leg,
        include: Union[str, Collection, None] = None
    ) -> Iterable[Measurement]:
        """
        Return available measurements.

        :param leg: Parent leg object
        :param include: Data types to include. By default all are included.
        """
        if isinstance(include, str):
            include = [include]
        elif include is not None:
            include = list(include)

        response = self._client.get(leg.uri + '/data',
                                    params={'include': include},
                                    stream=True)
        response.raise_for_status()
        it = response.iter_lines(decode_unicode=True)
        return (m for m in (Measurement(*line.split(',', maxsplit=2))
                            for line in it if line)
                if m.data)

    def get_raw_data(
        self,
        leg: Leg,
        include: Union[str, Collection, None] = None
    ) -> Iterable[str]:
        """
        Return the raw data. Format depends on source.

        :param leg: Parent leg object
        :param include: Data types to include. By default all are included.
        """
        if isinstance(include, str):
            include = [include]
        elif include is not None:
            include = list(include)

        response = self._client.get(leg.uri + '/raw',
                                    params={'include': include},
                                    stream=True)
        response.raise_for_status()
        return response.iter_lines(decode_unicode=True)



"""
Python library for interacting with the SRF's core data platform.

.. note::
    The ``SRFData`` implementation is designed for personal scripting use only.
    If you wish to use API access on behalf of others as part of some
    service, see the ``oauth`` module and contact SRF technical support.

Examples
--------
.. code:: python

    # initialise client for personal scripting use
    from srf_client import SRFData
    srf = SRFData(api_key='...')

    # use a file-based http cache
    from cachecontrol.caches import FileCache
    srf = SRFData(api_key='...', cache=FileCache('/var/cache/srf_client'))

    # get last recorded leg
    leg = srf.legs.find_all().last.items[-1]

    # plot ambient pressure for leg
    import pandas as pd
    df = pd.DataFrame(leg.get_data('13'))
    df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms')
    df['data'] = pd.to_numeric(df['data'], errors='coerce')
    df.plot('timestamp', 'data')

    # get vehicle and driver id
    vin = di = None
    for measurement in leg.get_data({'VIN', 'DI'}):
        if measurement.type == 'VIN':
            vin = measurement.data
        if measurement.type == 'DI':
            di = measurement.data.split(',')[0]
        if vin and di:
            break

    # list all DAF trucks
    from srf_client import prefix
    page = srf.vehicles.find_all(vin=prefix('XLR'))

    # plot total fuel consumption against total distance for recent trip
    from srf_client import contains
    trip = srf.trips.find_all(types=contains('LFC', 'VDHR')).last.items[0]
    df = trip.get_data_frame(['LFC', 'VDHR'], timedelta(minutes=1))
    df.plot.scatter('VDHR hr total vehicle distance',
                    'LFC engine total fuel used')

"""

import pkg_resources
__version__ = pkg_resources.require(__name__)[0].version

from .client import *
from .filter import *
from .model import *
from .sort import *

__all__ = client.__all__ + filter.__all__ + model.__all__ + sort.__all__

try:
    from . import pandas
except ImportError:
    pass

try:
    from .oauth import SRFDataOAuth
except ImportError:
    def SRFDataOAuth(**kwargs):
        """SRF Data client. Requires extra dependencies srf_client[oauth]."""
        raise RuntimeError('Required dependencies not installed')
__all__ += ['SRFDataOAuth']



from .model import Order, Route
from .paging import Page, PagingClient
from .util import parse_datetime, parse_distance


class OrdersClient(PagingClient[Order]):
    """
    Access to Orders API.

    Acquire an instance via ``SRFData.orders``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'orders')

    def _parse_obj(self, data, uri) -> Order:
        return Order(
            client=self._client,
            uri=uri,
            route_uri=data['route']['_location'],
            time_window_start=parse_datetime(data.get('timeWindowStart')),
            time_window_end=parse_datetime(data.get('timeWindowEnd')),
            pickup=data.get('pickup'),
            pickup_arrive=parse_datetime(data.get('pickupArrive')),
            pickup_depart=parse_datetime(data.get('pickupDepart')),
            destination=data.get('destination'),
            destination_arrive=parse_datetime(data.get('destinationArrive')),
            destination_depart=parse_datetime(data.get('destinationDepart')),
            ambient=data.get('ambient'),
            chilled=data.get('chilled'),
            frozen=data.get('frozen'),
            weight=data.get('weight'),
            distance=parse_distance(data.get('distance'))
        )

    def find_all(self, lazy=False, **kwargs) -> Page[Order]:
        """
        Find all known orders.

        :param lazy: Defer fetching of the Order objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/orders', params)

    def find_by_route(self, route: Route, lazy=False, **kwargs) -> Page[Order]:
        """
        Find all orders for the given route.

        :param route: Parent route object
        :param lazy: Defer fetching of the Leg objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page(route.uri + '/orders', params)



from typing import Optional

from .model import Vehicle
from .paging import Page, PagingClient


class VehiclesClient(PagingClient[Vehicle]):
    """
    Access to vehicles API.

    Acquire an instance via ``SRFData.vehicles``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'vehicles')

    def _parse_obj(self, data, uri) -> Vehicle:
        return Vehicle(
            client=self._client,
            uri=uri,
            registration=data.get('registration'),
            vin=data.get('vin'),
            calibrated=data.get('calibrated', False),
            make=data.get('make'),
            model=data.get('model'),
            type=data.get('type'),
            fuel=data.get('fuel'),
            weight_class=data.get('weightClass'),
            euro_standard=data.get('euroStandard'),
            description=data.get('description'),
            org_uri=data['organisation']['_location']
        )

    def get(self, *, uri=None, obj_id=None) -> Optional[Vehicle]:
        """
        Get a single vehicle by its URI or unique identifier.

        :param uri: Uniform resource locator
        :param obj_id: Vehicle registration, VIN, or internal ID.
        """
        return super().get(uri=uri, obj_id=obj_id)

    def find_all(self, lazy=False, **kwargs) -> Page[Vehicle]:
        """
        Find all known vehicles.

        :param lazy: Defer fetching of the Vehicle objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/vehicles', params)



from typing import Optional

from immutabledict import immutabledict

from .model import ExternalVehicleClass, VehicleClass
from .paging import Page, PagingClient


class VehicleClassesClient(PagingClient[VehicleClass]):
    """
    Access to VehicleClass API.

    Acquire an instance via ``SRFData.vehicle_classes``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'vehicle_classes')

    def _parse_obj(self, data, uri) -> VehicleClass:
        return VehicleClass(
            client=self._client,
            uri=uri,
            org_uri=data['organisation']['_location'],
            description=data.get('description'),
            make=data.get('make'),
            model=data.get('model'),
            type=data.get('type'),
            fuel=data.get('fuel'),
            fuel_capacity=data.get('fuelCapacity'),
            fuel_efficiency=data.get('fuelEfficiency'),
            weight_class=data.get('weightClass'),
            euro_standard=data.get('euroStandard'),
            drive_power=data.get('drivePower'),
            fridge_power=data.get('fridgePower'),
            limits=immutabledict(data.get('limits', {}))
        )

    def find_all(self, lazy=False, **kwargs) -> Page[VehicleClass]:
        """
        Find all known vehicle_classes.

        :param lazy: Defer fetching of the VehicleClass objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/vehicle_classes', params)

    def lookup(self, registration: str) -> Optional[ExternalVehicleClass]:
        """
        Lookup vehicle details from an external system, e.g. the DVLA.

        :param registration: Vehicle registration to find
        """
        response = self._client.get(f'/{self._fragment}/lookup',
                                    params={'registration': registration})
        if response.status_code == 404:
            return None
        response.raise_for_status()
        data = response.json()
        return ExternalVehicleClass(
            description=data.get('description'),
            make=data.get('make'),
            model=data.get('model'),
            type=data.get('type'),
            fuel=data.get('fuel'),
            fuel_capacity=data.get('fuelCapacity'),
            fuel_efficiency=data.get('fuelEfficiency'),
            weight_class=data.get('weightClass'),
            euro_standard=data.get('euroStandard'),
            drive_power=data.get('drivePower'),
            fridge_power=data.get('fridgePower')
        )



from .model import Trial
from .paging import Page, PagingClient


class TrialsClient(PagingClient[Trial]):
    """
    Access to Trials API.

    Acquire an instance via ``SRFData.trials``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'trials')

    def _parse_obj(self, data, uri) -> Trial:
        return Trial(
            client=self._client,
            uri=uri,
            description=data['description'],
            org_uri=data['organisation']['_location'],
            active=data.get('active')
        )

    def find_all(self, lazy=False, **kwargs) -> Page[Trial]:
        """
        Find all known trials.

        :param lazy: Defer fetching of the Trial objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/trials', params)



from abc import ABC, abstractmethod
from typing import Generic, Iterable, Optional, Sequence, TypeVar

import attr
import humps
from lazy_object_proxy import Proxy

T = TypeVar('T')


# https://github.com/python-attrs/attrs/issues/313
@attr.s(frozen=True, auto_attribs=True)
class Page(Generic[T]):
    """
    A page of results.

    :param items: Sequence of result items
    :param total: Total number of available items
    :param first: First page of results
    :param last: Last page of results
    :param next: Next page of results
    :param prev: Previous page of results
    """

    _client: 'PagingClient[T]'
    items: Sequence[T]
    total: int
    _first: str
    _last: str
    _next: Optional[str] = None
    _prev: Optional[str] = None

    @property
    def first(self) -> 'Page[T]':  # noqa: D102
        return self._client.get_page(self._first)

    @property
    def last(self) -> 'Page[T]':  # noqa: D102
        return self._client.get_page(self._last)

    @property
    def next(self) -> Optional['Page[T]']:  # noqa: D102
        if self._next:
            return self._client.get_page(self._next)
        else:
            return None

    @property
    def prev(self) -> Optional['Page[T]']:  # noqa: D102
        if self._prev:
            return self._client.get_page(self._prev)
        else:
            return None


class PagingClient(Generic[T], ABC):
    """
    Base class for paging support.

    :param page_size: the default page size
    """

    page_size = 100

    @staticmethod
    def _filter_params(**kwargs):
        """Remove reserved query params."""
        return {k: v for k, v in kwargs.items()
                if k not in ('includeItems', 'page', 'size')}

    def __init__(self, client, fragment):
        """
        Construct base class.

        :param client: root SRFData client
        :param fragment: URI path fragment
        """
        from . import SRFData
        self._client: SRFData = client
        self._fragment = fragment

    @abstractmethod
    def _parse_obj(self, data, uri) -> T:
        ...

    def _parse_item(self, data) -> T:
        try:
            return self._parse_obj(data['_item'], data['_location'])
        except KeyError:
            return Proxy(lambda: self.get(uri=data['_location']))

    def get(self, *, uri=None, obj_id=None) -> Optional[T]:
        """
        Get a single object by its URI or ID.

        In general IDs are not exposed to clients.
        The ``find_*`` methods should be used instead.
        """
        if bool(uri) == bool(obj_id):
            raise ValueError('Exactly one of uri or obj_id must be given')
        if obj_id:
            uri = f'/{self._fragment}/{obj_id}'

        response = self._client.get(uri)
        if response.status_code == 404:
            return None
        response.raise_for_status()
        return self._parse_obj(response.json(), response.url)

    def get_page(self, uri, params=None) -> Page[T]:
        """
        Get a page of objects.

        In general this method should not be used directly.
        Access is via ``find_*`` methods and attributes of ``Page``.
        """
        if params is None:
            params = {'size': self.page_size}
        elif 'size' not in params:
            params['size'] = self.page_size

        response = self._client.get(uri, params=humps.camelize(params))
        response.raise_for_status()
        data = response.json()
        return Page(
            client=self,
            items=tuple(self._parse_item(i) for i in data['_items']),
            total=data['_total'],
            first=data['_first'],
            last=data['_last'],
            next=data.get('_next'),
            prev=data.get('_prev')
        )


def paged_items(page: Page[T]) -> Iterable[T]:
    """
    Iterate over all items in all further pages.

    :param page: Page to start from
    """
    while page:
        for item in page.items:
            yield item
        page = page.next



#!/usr/bin/env python

import os

from setuptools import find_packages, setup

__version__ = '1.1'

try:
    if os.environ['GIT_BRANCH'] == 'master':
        __version__ += '.dev' + os.environ['BUILD_NUMBER']
except KeyError:
    pass

setup(
    name='srf-client',
    description='Centre for Sustainable Road Freight - Data Platform client',
    version=__version__,
    packages=find_packages(),
    license='MIT',
    author='James Howe',
    author_email='jmh205@cam.ac.uk',
    maintainer='Centre for Sustainable Road Freight',
    maintainer_email='tech@csrf.ac.uk',
    url='https://data.csrf.ac.uk',
    project_urls={
        'Documentation': 'https://data.csrf.ac.uk/python/docs'
    },
    python_requires='>=3.7',
    install_requires=[
        'attrs >=19.1',
        'CacheControl',
        'immutabledict <2',
        'geopy <3',
        'iso8601',
        'lazy-object-proxy >=1.4, <2',
        'pyhumps >=1.6, <2',
        'requests >=2.22, <3'
    ],
    extras_require={
        'oauth': ['requests-oauthlib <2', 'oauthlib'],
        'pandas': ['numpy', 'pandas>=1.2']
    },
    setup_requires=[
        'pytest-runner',
        'flake8 <4',
        'sphinx <4',
        'wheel'
    ],
    tests_require=[
        'flake8-docstrings',
        'pydocstyle <4',
        'pytest >=6.2.4, <7',
        'pytest-cov',
        'pytest-flake8',
        'pytest-mock',
        'requests-mock >=1.9.0',
        'requests-oauthlib'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Natural Language :: English'
    ]
)



from .model import Fleet, FleetEntry
from .paging import Page, paged_items, PagingClient


class FleetsClient(PagingClient[Fleet]):
    """
    Access to Fleets API.

    Acquire an instance via ``SRFData.fleets``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'fleets')
        self._entries = FleetEntriesClient(client)

    def _parse_obj(self, data, uri) -> Fleet:
        return Fleet(
            client=self._client,
            uri=uri,
            org_uri=data['organisation']['_location'],
            name=data['name'],
            entries=tuple(paged_items(self._entries.find_by_fleet(uri)))
        )

    def find_all(self, lazy=False, **kwargs) -> Page[Fleet]:
        """
        Find all known fleets.

        :param lazy: Defer fetching of the Fleet objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/fleets', params)


class FleetEntriesClient(PagingClient[FleetEntry]):
    """Internal access to Fleet Entries API."""

    def __init__(self, client):  # noqa: D107
        super().__init__(client, None)

    def _parse_obj(self, data, uri) -> FleetEntry:
        return FleetEntry(
            client=self._client,
            location_uri=data['location']['_location'],
            vehicle_class_uri=data['vehicleClass']['_location'],
            count=data['count']
        )

    def _parse_item(self, data) -> FleetEntry:
        return self._parse_obj(data['_item'], NotImplemented)

    def find_by_fleet(self, fleet_uri: str) -> Page[FleetEntry]:
        """
        Get fleet composition entries.

        :param fleet_uri: Fleet to enumerate
        """
        return self.get_page(fleet_uri + '/entries')



"""Core HTTP client."""
from collections import OrderedDict

import logging
from cachecontrol import CacheControl
from geopy import Point
from requests import RequestException, Response, Session
from requests.auth import AuthBase
from requests.utils import default_user_agent
from typing import Iterable, Mapping, Optional
from urllib.parse import urljoin

from . import (
    __version__, fleets, legs, locations, orders, organisations, routes,
    trailers, trials, trips, vehicle_classes, vehicles,
)

__all__ = ['SRFData']


class ApiKeyAuth(AuthBase):
    def __init__(self, api_key):
        self.api_key = api_key

    def __call__(self, r):
        r.headers['Authorization'] = 'Bearer ' + self.api_key
        return r


class SRFData:
    """
    SRF Data client. Provides basic access to the REST API.

    Access to objects is via typed attributes.
    Most users should only need the ``find_*`` methods.

    :param fleets: Access to fleets API.
    :param legs: Access to legs API.
    :param locations: Access to locations API.
    :param orders: Access to orders API.
    :param organisations: Access to organisations API.
    :param routes: Access to routes API.
    :param trailers: Access to trailers API.
    :param trials: Access to trials API.
    :param trips: Access to trips API.
    :param vehicle_classes: Access to vehicles classes API.
    :param vehicles: Access to vehicles API.
    """

    from .model import DataType
    _log = logging.getLogger('srf_client.SRFData')

    def __init__(self, root='https://data.csrf.ac.uk/api/', cache=None,
                 **kwargs):
        """
        Initialise a new instance.

        :param api_key: Personal authentication token for the API
        :param root: (optional) API root URL
        :param cache: (optional) provide an alternate cache implementation,
            or ``False`` to disable the cache
        """
        self._root = root
        if self._root[-1] != '/':
            self._root += '/'

        self._make_session(**kwargs)
        if cache is not False:
            self._session = CacheControl(self._session, cache=cache)
        self._session.verify = True
        self._session.headers.update({
            'Accept': 'application/json, */*;q=0.8',
            'User-Agent': '{}/{} {}'.format('srf-client', __version__,
                                            default_user_agent())
        })

        self.fleets = fleets.FleetsClient(self)
        self.legs = legs.LegsClient(self)
        self.locations = locations.LocationsClient(self)
        self.orders = orders.OrdersClient(self)
        self.organisations = organisations.OrganisationsClient(self)
        self.routes = routes.RoutesClient(self)
        self.trailers = trailers.TrailersClient(self)
        self.trials = trials.TrialsClient(self)
        self.trips = trips.TripsClient(self)
        self.vehicle_classes = vehicle_classes.VehicleClassesClient(self)
        self.vehicles = vehicles.VehiclesClient(self)

    def _make_session(self, **kwargs):
        self._session = Session()
        self._session.auth = ApiKeyAuth(kwargs['api_key'])

    def _prepare_request(self, url, kwargs):
        if url[0] == '/':  # avoid urljoin confusion with leading slash
            url = url[1:]
        url = urljoin(self._root, url)
        if not url.startswith(self._root):
            raise ValueError('Resolved url not part of the configured API')
        if 'timeout' not in kwargs:
            kwargs['timeout'] = (10, 60)
        return url, kwargs

    def get(self, url, **kwargs) -> Response:
        """
        Perform a raw GET request.

        :param url: Request URL. Will be resolved relative to the API root.
        :param kwargs: Additional options for request
        """
        url, kwargs = self._prepare_request(url, kwargs)
        return self._session.get(url=url, **kwargs)

    def ping(self, timeout=2) -> bool:
        """
        Check whether the service is reachable.

        May fail due to server-side errors, invalid authentication,
        or incorrect configuration.

        :param timeout: Time in seconds to wait for a response.
        """
        try:
            response = self.get('/', timeout=timeout)
        except RequestException:
            self._log.warning('Exception during ping()', exc_info=True)
            return False
        return response.headers['Server'].startswith('data-platform/') \
            and response.status_code in (200, 404)

    def get_types(self, raw: bool = False) -> Mapping[str, 'DataType']:
        """
        Get the ``type`` definitions for the ``get_data`` measurements.

        :param raw: Return the types for ``get_raw_data`` instead..
        """
        from .model import DataType, FieldDef
        response = self.get('/types', params={'raw': raw})

        def parse(data_type):
            return DataType(
                type=data_type['type'],
                description=data_type['description'],
                fields=None if raw else tuple(
                    FieldDef(**f) for f in data_type['fields']
                )
            )
        return OrderedDict((t['type'], parse(t)) for t in response.json())

    def get_elevation(self, points: Iterable[Point]
                      ) -> Iterable[Optional[float]]:
        """
        Lookup elevation for given points.

        :param points: (latitude, longitude) points to query
        :return: iterable of corresponding elevation in metres
        """
        url, kwargs = self._prepare_request('/elevation', {})
        response = self._session.post(
            url,
            headers={'Content-Type': 'text/csv'},
            data=(f'{p[0]:.6f},{p[1]:.6f}\r\n'.encode('us-ascii')
                  for p in points),
            stream=True,
            **kwargs)
        response.raise_for_status()
        return (None if s == b'null' else float(s)
                for s in response.iter_lines())



from .model import Location
from .paging import Page, PagingClient
from .util import parse_point


class LocationsClient(PagingClient[Location]):
    """
    Access to Locations API.

    Acquire an instance via ``SRFData.locations``.
    """

    def __init__(self, client):  # noqa: D107
        super().__init__(client, 'locations')

    def _parse_obj(self, data, uri) -> Location:
        return Location(
            client=self._client,
            uri=uri,
            org_uri=data['organisation']['_location'],
            name=data.get('name'),
            post_code=data.get('postCode'),
            point=parse_point(data.get('point'))
        )

    def find_all(self, lazy=False, **kwargs) -> Page[Location]:
        """
        Find all known locations.

        :param lazy: Defer fetching of the Location objects
        :param kwargs: Additional field filters
        """
        params = {'includeItems': not lazy, **self._filter_params(**kwargs)}
        return self.get_page('/locations', params)
